@@ -29,46 +29,111 @@ import os
 import shutil
 import subprocess
 import pkg_resources
-import shutil
 import stat
 import tempfile
 from supervisor import xmlrpc
 import xmlrpclib
 import pwd
-from utils import launchBuildout, getCleanEnvironment,\
-                   dropPrivileges, bootstrapBuildout, updateFile,\
-                   getSoftwareUrlHash, SlapPopen
+import utils
 from svcbackend import getSupervisorRPC
 from exception import BuildoutFailedError, WrongPermissionError, \
     PathDoesNotExistError
+from networkcache import download_network_cached, upload_network_cached
+import tarfile
 
 REQUIRED_COMPUTER_PARTITION_PERMISSION = '0750'
 
 
 class Software(object):
   """This class is responsible of installing a software release"""
-  def __init__(self, url, software_root, console, buildout):
+  def __init__(self, url, software_root, console, buildout,
+      signature_private_key_file=None, signature_certificate_list=None,
+      upload_cache_url=None, upload_dir_url=None, shacache_cert_file=None,
+      shacache_key_file=None, shadir_cert_file=None, shadir_key_file=None,
+      download_binary_cache_url=None, upload_binary_cache_url=None,
+      download_binary_dir_url=None, upload_binary_dir_url=None):
     """Initialisation of class parameters
     """
     self.url = url
     self.software_root = software_root
+    self.software_url_hash = utils.getSoftwareUrlHash(self.url)
     self.software_path = os.path.join(self.software_root,
-                                      getSoftwareUrlHash(self.url))
+                                      self.software_url_hash)
     self.buildout = buildout
     self.logger = logging.getLogger('BuildoutManager')
     self.console = console
+    self.signature_private_key_file = signature_private_key_file
+    self.signature_certificate_list = signature_certificate_list
+    self.upload_cache_url = upload_cache_url
+    self.upload_dir_url = upload_dir_url
+    self.shacache_cert_file = shacache_cert_file
+    self.shacache_key_file = shacache_key_file
+    self.shadir_cert_file = shadir_cert_file
+    self.shadir_key_file = shadir_key_file
+    self.download_binary_cache_url = download_binary_cache_url
+    self.upload_binary_cache_url = upload_binary_cache_url
+    self.download_binary_dir_url = download_binary_dir_url
+    self.upload_binary_dir_url = upload_binary_dir_url
 
   def install(self):
+    """ Fetches binary cache if possible.
+    Installs from buildout otherwise.
+    """
+    tarname = self.software_url_hash
+    cache_dir = tempfile.mkdtemp()
+    tarpath = os.path.join(cache_dir, tarname)
+    if (not os.path.exists(self.software_path)) \
+      and download_network_cached(
+          self.download_binary_cache_url,
+          self.download_binary_dir_url,
+          self.url, self.software_root,
+          self.software_url_hash,
+          tarpath, self.logger,
+          self.signature_certificate_list):
+        tar = tarfile.open(tarpath)
+        try:
+          self.logger.info("Extracting archive of cached software release...")
+          tar.extractall(path=self.software_root)
+        finally:
+          tar.close()
+    else:
+        self._install_from_buildout()
+        if (self.software_root and self.url and self.software_url_hash \
+                               and self.upload_binary_cache_url \
+                               and self.upload_binary_dir_url):
+          self.logger.info("Creating archive of software release...")
+          tar = tarfile.open(tarpath, "w:gz")
+          try:
+            tar.add(self.software_path, arcname=self.software_url_hash)
+          finally:
+            tar.close()
+          self.logger.info("Trying to upload archive of software release...")
+          upload_network_cached(
+            self.software_root,
+            self.url, self.software_url_hash,
+            self.upload_binary_cache_url,
+            self.upload_binary_dir_url,
+            tarpath, self.logger,
+            self.signature_private_key_file,
+            self.shacache_cert_file,
+            self.shacache_key_file,
+            self.shadir_cert_file,
+            self.shadir_key_file)
+    shutil.rmtree(cache_dir)
+      
+  def _install_from_buildout(self):
     """ Fetches buildout configuration from the server, run buildout with
     it. If it fails, we notify the server.
     """
     self.logger.info("Installing software release %s..." % self.url)
+    root_stat_info = os.stat(self.software_root)
+    os.environ = utils.getCleanEnvironment(pwd.getpwuid(root_stat_info.st_uid
+      ).pw_dir)
     if not os.path.isdir(self.software_path):
       os.mkdir(self.software_path)
     extends_cache = tempfile.mkdtemp()
     if os.getuid() == 0:
       # In case when running as root copy ownership, to simplify logic
-      root_stat_info = os.stat(self.software_root)
       for path in [self.software_path, extends_cache]:
         path_stat_info = os.stat(path)
         if root_stat_info.st_uid != path_stat_info.st_uid or\
@@ -78,23 +143,52 @@ class Software(object):
     try:
       buildout_parameter_list = [
         'buildout:extends-cache=%s' % extends_cache,
-        'buildout:directory=%s' % self.software_path,
-        '-c', self.url]
-      bootstrapBuildout(self.software_path, self.buildout,
+        'buildout:directory=%s' % self.software_path,]
+
+      if self.signature_private_key_file or \
+          self.upload_cache_url or \
+            self.upload_dir_url is not None:
+        buildout_parameter_list.append('buildout:networkcache-section=networkcache')
+      for  buildout_option, value in (
+         ('%ssignature-private-key-file=%s', self.signature_private_key_file),
+         ('%supload-cache-url=%s', self.upload_cache_url),
+         ('%supload-dir-url=%s', self.upload_dir_url),
+         ('%sshacache-cert-file=%s', self.shacache_cert_file),
+         ('%sshacache-key-file=%s', self.shacache_key_file),
+         ('%sshadir-cert-file=%s', self.shadir_cert_file),
+         ('%sshadir-key-file=%s', self.shadir_key_file),
+         ):
+        if value:
+          buildout_parameter_list.append( \
+              buildout_option % ('networkcache:', value))
+
+      buildout_parameter_list.extend(['-c', self.url])
+      utils.bootstrapBuildout(self.software_path, self.buildout,
           additional_buildout_parametr_list=buildout_parameter_list,
           console=self.console)
-      launchBuildout(self.software_path,
+      utils.launchBuildout(self.software_path,
                      os.path.join(self.software_path, 'bin', 'buildout'),
                      additional_buildout_parametr_list=buildout_parameter_list,
                      console=self.console)
     finally:
       shutil.rmtree(extends_cache)
 
-  def remove(self):
-    """Removes the part that was installed.
-    """
+  def destroy(self):
+    """Removes software release."""
+    def retry(func, path, exc):
+      # inspired on slapos.buildout hard remover
+      if func == os.path.islink:
+        os.unlink(path)
+      else:
+        os.chmod (path, 0600)
+        func(path)
     try:
-      shutil.rmtree(self.software_path)
+      if os.path.exists(self.software_path):
+        self.logger.info('Removing path %r' % self.software_path)
+        shutil.rmtree(self.software_path, onerror=retry)
+      else:
+        self.logger.info('Path %r does not exists, no need to remove.' %
+            self.software_path)
     except IOError as error:
       error_string = "I/O error while removing software (%s): %s" % (self.url,
                                                                      error)
@@ -175,13 +269,16 @@ class Partition(object):
     if not os.path.isdir(self.instance_path):
       raise PathDoesNotExistError('Please create partition directory %s'
                                            % self.instance_path)
-    permission = oct(stat.S_IMODE(os.stat(self.instance_path).st_mode))
+    instance_stat_info = os.stat(self.instance_path)
+    permission = oct(stat.S_IMODE(instance_stat_info.st_mode))
     if permission != REQUIRED_COMPUTER_PARTITION_PERMISSION:
       raise WrongPermissionError('Wrong permissions in %s : actual ' \
                                           'permissions are : %s, wanted ' \
                                           'are %s' %
                                           (self.instance_path, permission,
                                             REQUIRED_COMPUTER_PARTITION_PERMISSION))
+    os.environ = utils.getCleanEnvironment(pwd.getpwuid(
+      instance_stat_info.st_uid).pw_dir)
     # Generates buildout part from template
     # TODO how to fetch the good template? Naming conventions?
     template_location = os.path.join(self.software_path, 'template.cfg')
@@ -237,9 +334,9 @@ class Partition(object):
       kw = dict()
       if not self.console:
         kw.update(stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
-      process_handler = SlapPopen(invocation_list,
-        preexec_fn=lambda: dropPrivileges(uid, gid), cwd=self.instance_path,
-        env=getCleanEnvironment(pwd.getpwuid(uid).pw_dir), **kw)
+      process_handler = utils.SlapPopen(invocation_list,
+        preexec_fn=lambda: utils.dropPrivileges(uid, gid), cwd=self.instance_path,
+        env=utils.getCleanEnvironment(pwd.getpwuid(uid).pw_dir), **kw)
       result_std = process_handler.communicate()[0]
       if self.console:
         result_std = 'Please consult messages above.'
@@ -251,12 +348,12 @@ class Partition(object):
 
     if not os.path.exists(buildout_binary):
       # use own buildout generation
-      bootstrapBuildout(self.instance_path, self.buildout,
+      utils.bootstrapBuildout(self.instance_path, self.buildout,
         ['buildout:bin-directory=%s'% os.path.join(self.instance_path,
         'sbin')], console=self.console)
       buildout_binary = os.path.join(self.instance_path, 'sbin', 'buildout')
     # Launches buildout
-    launchBuildout(self.instance_path,
+    utils.launchBuildout(self.instance_path,
                    buildout_binary, console=self.console)
     # Generates supervisord configuration file from template
     self.logger.info("Generating supervisord config file from template...")
@@ -296,7 +393,7 @@ class Partition(object):
           HOME=pwd.getpwuid(uid).pw_dir,
           USER=pwd.getpwuid(uid).pw_name,
         )
-      updateFile(self.supervisord_partition_configuration_path,
+      utils.updateFile(self.supervisord_partition_configuration_path,
           partition_supervisor_configuration)
     self.updateSupervisor()
 
@@ -306,8 +403,14 @@ class Partition(object):
     """
     supervisor = self.getSupervisorRPC()
     partition_id = self.computer_partition.getId()
-    supervisor.startProcessGroup(partition_id, False)
-    self.logger.info("Requested start of %s..." % self.computer_partition.getId())
+    try:
+      supervisor.startProcessGroup(partition_id, False)
+    except xmlrpclib.Fault, e:
+      if e.faultString.startswith('BAD_NAME:'):
+        self.logger.info("Nothing to start on %s..." % \
+                         self.computer_partition.getId())
+    else:
+      self.logger.info("Requested start of %s..." % self.computer_partition.getId())
 
   def stop(self):
     """Asks supervisord to stop the instance."""
@@ -343,9 +446,9 @@ class Partition(object):
       kw = dict()
       if not self.console:
         kw.update(stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
-      process_handler = SlapPopen([destroy_executable_location],
-        preexec_fn=lambda: dropPrivileges(uid, gid), cwd=self.instance_path,
-        env=getCleanEnvironment(pwd.getpwuid(uid).pw_dir), **kw)
+      process_handler = utils.SlapPopen([destroy_executable_location],
+        preexec_fn=lambda: utils.dropPrivileges(uid, gid), cwd=self.instance_path,
+        env=utils.getCleanEnvironment(pwd.getpwuid(uid).pw_dir), **kw)
       result_std = process_handler.communicate()[0]
       if self.console:
         result_std = 'Please consult messages above'
