@@ -35,6 +35,7 @@ if sys.version_info < (2, 6):
 import socket
 import subprocess
 import traceback
+import time
 #from time import strftime
 
 from SlapObject import Software, Partition, WrongPermissionError, \
@@ -48,12 +49,16 @@ from utils import setFinished
 from utils import getSoftwareUrlHash
 from slapos import slap
 from slapos.slap import NotFoundError
+from slapos.slap.slap import ServerError
 from utils import dropPrivileges
 from utils import SlapPopen
 from svcbackend import launchSupervisord
 import tempfile
+from time import strftime
 import StringIO
 from lxml import etree
+from time import sleep
+from random import random
 
 
 MANDATORY_PARAMETER_LIST = [
@@ -97,6 +102,7 @@ def parseArgumentTupleAndReturnSlapgridObject(*argument_tuple):
   parser.add_argument("--key_file", help="SSL Authorisation key file.")
   parser.add_argument("--cert_file",
       help="SSL Authorisation certificate file.")
+  parser.add_argument("--signature_private_key_file", help="Signature private key file.")
   parser.add_argument("--master_ca_file", help="Root certificate of SlapOS "
       "master key.")
   parser.add_argument("--certificate_repository_path",
@@ -105,8 +111,13 @@ def parseArgumentTupleAndReturnSlapgridObject(*argument_tuple):
       help="Enables console output and live output from subcommands.")
   parser.add_argument("-v", "--verbose", action="store_true", default=False,
       help="Be verbose.")
+  parser.add_argument("--promise-timeout",
+                      type=int, default=3,
+                      help="Promise timeout in seconds.")
   parser.add_argument("configuration_file", nargs=1, type=argparse.FileType(),
       help="SlapOS configuration file.")
+  parser.add_argument("--maximal_delay", help="The maximal delay value in seconds. " \
+                    "A negative value leads start immediately.")
 
   # Parses arguments
   if argument_tuple == ():
@@ -123,6 +134,8 @@ def parseArgumentTupleAndReturnSlapgridObject(*argument_tuple):
   slapgrid_configuration.readfp(configuration_file)
   # Merges the two dictionnaries
   option_dict = dict(slapgrid_configuration.items("slapos"))
+  if slapgrid_configuration.has_section("networkcache"):
+    option_dict.update(dict(slapgrid_configuration.items("networkcache")))
   for argument_key, argument_value in vars(argument_option_instance
       ).iteritems():
     if argument_value is not None:
@@ -164,8 +177,19 @@ def parseArgumentTupleAndReturnSlapgridObject(*argument_tuple):
   key_file = option_dict.get('key_file')
   cert_file = option_dict.get('cert_file')
   master_ca_file = option_dict.get('master_ca_file')
+  signature_private_key_file = option_dict.get('signature_private_key_file')
 
-  for f in [key_file, cert_file, master_ca_file]:
+  mandatory_file_list = [key_file, cert_file, master_ca_file]
+  # signature_private_key_file is not mandatory, we must be able to run
+  # slapgrid scripts without this parameter.
+  if signature_private_key_file:
+    mandatory_file_list.append(signature_private_key_file)
+
+  for k in ['shacache-cert-file', 'shacache-key-file', 'shadir-cert-file',
+      'shadir-key-file']:
+    mandatory_file_list.append(option_dict.get(k, None))
+
+  for f in mandatory_file_list:
     if f is not None:
       if not os.path.exists(f):
         parser.error('File %r does not exists.' % f)
@@ -184,6 +208,24 @@ def parseArgumentTupleAndReturnSlapgridObject(*argument_tuple):
   if not option_dict.get('supervisord_socket'):
     option_dict['supervisord_socket'] = \
       os.path.join(option_dict['instance_root'], 'supervisord.socket')
+
+  signature_certificate_list_string = \
+    option_dict.get('signature-certificate-list', None)
+  if signature_certificate_list_string is not None:
+    cert_marker = "-----BEGIN CERTIFICATE-----"
+    signature_certificate_list = [cert_marker + '\n' + q.strip() \
+      for q in signature_certificate_list_string.split(cert_marker) \
+        if q.strip()]
+  else:
+    signature_certificate_list = None
+
+  maximal_delay = float(option_dict.get("maximal_delay", "300"))
+  if maximal_delay > 0:
+    duration = maximal_delay * random()
+    logging.info("Sleeping for %s seconds. To disable this feature, " \
+                    "check maximal_delay parameter in manual." % duration)
+    time.sleep(duration)
+
   # Returning new Slapgrid instance and options
   return ([Slapgrid(software_root=option_dict['software_root'],
             instance_root=option_dict['instance_root'],
@@ -197,8 +239,26 @@ def parseArgumentTupleAndReturnSlapgridObject(*argument_tuple):
             cert_file=cert_file,
             master_ca_file=master_ca_file,
             certificate_repository_path=certificate_repository_path,
+            signature_private_key_file=signature_private_key_file,
+            signature_certificate_list=signature_certificate_list,
+            download_binary_cache_url=\
+              option_dict.get('download-binary-cache-url', None),
+            upload_binary_cache_url=\
+              option_dict.get('upload-binary-cache-url', None),
+            upload_cache_url=option_dict.get('upload-cache-url', None),
+            download_binary_dir_url=\
+              option_dict.get('download-binary-dir-url', None),
+            upload_binary_dir_url=\
+              option_dict.get('upload-binary-dir-url', None),
+            upload_dir_url=option_dict.get('upload-dir-url', None),
             console=option_dict['console'],
-            buildout=option_dict.get('buildout')),
+            buildout=option_dict.get('buildout'),
+            promise_timeout=option_dict['promise_timeout'],
+            shacache_cert_file=option_dict.get('shacache-cert-file', None),
+            shacache_key_file=option_dict.get('shacache-key-file', None),
+            shadir_cert_file=option_dict.get('shadir-cert-file', None),
+            shadir_key_file=option_dict.get('shadir-key-file', None),
+            ),
           option_dict])
 
 
@@ -254,6 +314,10 @@ class Slapgrid(object):
   """ Main class for SlapGrid. Fetches and processes informations from master
   server and pushes usage information to master server.
   """
+
+  class PromiseError(Exception):
+    pass
+
   def __init__(self,
                software_root,
                instance_root,
@@ -265,9 +329,22 @@ class Slapgrid(object):
                buildout,
                key_file=None,
                cert_file=None,
+               signature_private_key_file=None,
+               signature_certificate_list=None,
+               download_binary_cache_url=None,
+               upload_binary_cache_url=None,
+               upload_cache_url=None,
+               download_binary_dir_url=None,
+               upload_binary_dir_url=None,
+               upload_dir_url=None,
                master_ca_file=None,
                certificate_repository_path=None,
-               console=False):
+               console=False,
+               promise_timeout=3,
+               shacache_cert_file=None,
+               shacache_key_file=None,
+               shadir_cert_file=None,
+               shadir_key_file=None):
     """Makes easy initialisation of class parameters"""
     # Parses arguments
     self.software_root = os.path.abspath(software_root)
@@ -281,6 +358,18 @@ class Slapgrid(object):
     self.cert_file = cert_file
     self.master_ca_file = master_ca_file
     self.certificate_repository_path = certificate_repository_path
+    self.signature_private_key_file = signature_private_key_file
+    self.signature_certificate_list = signature_certificate_list
+    self.download_binary_cache_url = download_binary_cache_url
+    self.upload_binary_cache_url = upload_binary_cache_url
+    self.upload_cache_url = upload_cache_url
+    self.download_binary_dir_url = download_binary_dir_url
+    self.upload_binary_dir_url = upload_binary_dir_url
+    self.upload_dir_url = upload_dir_url
+    self.shacache_cert_file = shacache_cert_file
+    self.shacache_key_file = shacache_key_file
+    self.shadir_cert_file = shadir_cert_file
+    self.shadir_key_file = shadir_key_file
     # Configures logger
     self.logger = logging.getLogger('Slapgrid')
     # Creates objects from slap module
@@ -294,6 +383,7 @@ class Slapgrid(object):
         os.path.join(self.instance_etc_directory, 'supervisord.conf.d')
     self.console = console
     self.buildout = buildout
+    self.promise_timeout = promise_timeout
 
   def checkEnvironmentAndCreateStructure(self):
     """Checks for software_root and instance_root existence, then creates
@@ -337,7 +427,7 @@ class Slapgrid(object):
       computer_partition_list = self.computer.getComputerPartitionList()
     except socket.error as error:
       self.logger.fatal(error)
-      sys.exit(1)
+      raise
     return computer_partition_list
 
   def processSoftwareReleaseList(self):
@@ -348,11 +438,31 @@ class Slapgrid(object):
     logger.info("Processing software releases...")
     clean_run = True
     for software_release in self.computer.getSoftwareReleaseList():
+      state = software_release.getState()
       try:
-        software_release.building()
         software_release_uri = software_release.getURI()
-        Software(url=software_release_uri, software_root=self.software_root,
-            console=self.console, buildout=self.buildout).install()
+        software = Software(url=software_release_uri,
+            software_root=self.software_root,
+            console=self.console, buildout=self.buildout,
+            signature_private_key_file=self.signature_private_key_file,
+            signature_certificate_list=self.signature_certificate_list,
+            download_binary_cache_url=self.download_binary_cache_url,
+            upload_binary_cache_url=self.upload_binary_cache_url,
+            upload_cache_url=self.upload_cache_url,
+            download_binary_dir_url=self.download_binary_dir_url,
+            upload_binary_dir_url=self.upload_binary_dir_url,
+            upload_dir_url=self.upload_dir_url,
+            shacache_cert_file=self.shacache_cert_file,
+            shacache_key_file=self.shacache_key_file,
+            shadir_cert_file=self.shadir_cert_file,
+            shadir_key_file=self.shadir_key_file)
+        if state == 'available':
+          software_release.building()
+          software.install()
+        elif state == 'destroyed':
+          logger.info('Destroying %r...' % software_release_uri)
+          software.destroy()
+          logger.info('Destroyed %r.' % software_release_uri)
       except (SystemExit, KeyboardInterrupt):
         exception = traceback.format_exc()
         software_release.error(exception)
@@ -363,7 +473,10 @@ class Slapgrid(object):
         software_release.error(exception)
         clean_run = False
       else:
-        software_release.available()
+        if state == 'available':
+          software_release.available()
+        elif state == 'destroyed':
+          software_release.destroyed()
     logger.info("Finished software releases...")
     return clean_run
 
@@ -371,6 +484,59 @@ class Slapgrid(object):
     launchSupervisord(self.supervisord_socket,
         self.supervisord_configuration_path)
 
+  def _checkPromises(self, computer_partition):
+    self.logger.info("Checking promises...")
+    instance_path = os.path.join(self.instance_root,
+        computer_partition.getId())
+
+    uid, gid = None, None
+    stat_info = os.stat(instance_path)
+
+    #stat sys call to get statistics informations
+    uid = stat_info.st_uid
+    gid = stat_info.st_gid
+
+    promise_present = False
+    # Get the list of promises
+    promise_dir = os.path.join(instance_path, 'etc', 'promise')
+    if os.path.exists(promise_dir) and os.path.isdir(promise_dir):
+      cwd = instance_path
+      promises_list = os.listdir(promise_dir)
+
+      # Check whether every promise is kept
+      for promise in promises_list:
+        promise_present = True
+
+        command = [os.path.join(promise_dir, promise)]
+
+        promise = os.path.basename(command[0])
+        self.logger.info("Checking promise %r.", promise)
+
+        kw = dict(stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+
+        process_handler = SlapPopen(command,
+          preexec_fn=lambda: dropPrivileges(uid, gid),
+          cwd=cwd,
+          env={}, **kw)
+
+        time.sleep(self.promise_timeout)
+
+
+        if process_handler.poll() is None:
+          process_handler.terminate()
+          raise Slapgrid.PromiseError("The promise %r timed out" % promise)
+        elif process_handler.poll() != 0:
+          stderr = process_handler.communicate()[1]
+          if stderr is None:
+            stderr = 'No error output from %r.' % promise
+          else:
+            stderr = 'Promise %r:' % promise + stderr
+          raise Slapgrid.PromiseError(stderr)
+
+    if not promise_present:
+      self.logger.info("No promise.")
+
+
   def processComputerPartitionList(self):
     """Will start supervisord and process each Computer Partition.
     """
@@ -411,6 +577,7 @@ class Slapgrid(object):
           local_partition.install()
           computer_partition.available()
           local_partition.start()
+          self._checkPromises(computer_partition)
           computer_partition.started()
         elif computer_partition_state == "stopped":
           local_partition.install()
@@ -452,25 +619,89 @@ class Slapgrid(object):
         exception = traceback.format_exc()
         logger.error(exception)
         computer_partition.error(exception)
+
     logger.info("Finished computer partitions...")
     return clean_run
 
-  def validateXML(self, to_be_validated):
-    """Will validate the xml file"""
+  def validateXML(self, to_be_validated, xsd_model):
+    """Validates a given xml file"""
 
-    #We get the xsd model
-    fed = open('choose_the_path', 'r')
-    model = StringIO.StringIO(fed.read())
-    xmlschema_doc = etree.parse(model)
+    logger = logging.getLogger('XMLValidating')
+
+    #We retrieve the xsd model
+    xsd_model = StringIO.StringIO(xsd_model)
+    xmlschema_doc = etree.parse(xsd_model)
     xmlschema = etree.XMLSchema(xmlschema_doc)
 
     string_to_validate = StringIO.StringIO(to_be_validated)
-    document = etree.parse(string_to_validate)
+
+    try:
+      document = etree.parse(string_to_validate)
+    except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
+      logger.info('Failed to parse this XML report :  %s\n%s' % \
+        (to_be_validated, _formatXMLError(e)))
+      logger.error(_formatXMLError(e))
+      return False
+
     if xmlschema.validate(document):
       return True
 
     return False
 
+  def asXML(self, computer_partition_usage_list):
+    """Generates a XML report from computer partition usage list
+    """
+    xml_head = ""
+    xml_movements = ""
+    xml_foot = ""
+
+    xml_head = "<?xml version='1.0' encoding='utf-8'?>" \
+               "<journal>" \
+               "<transaction type=\"Sale Packing List\">" \
+               "<title>Resource consumptions</title>" \
+               "<start_date></start_date>" \
+               "<stop_date>%s</stop_date>" \
+               "<reference>%s</reference>" \
+               "<currency></currency>" \
+               "<payment_mode></payment_mode>" \
+               "<category></category>" \
+               "<arrow type=\"Administration\">" \
+               "<source></source>" \
+               "<destination></destination>" \
+               "</arrow>" \
+               % (strftime("%Y-%m-%d at %H:%M:%S"), 
+                  self.computer_id)
+
+    for computer_partition_usage in computer_partition_usage_list:
+      try:
+        usage_string = StringIO.StringIO(computer_partition_usage.usage)
+        root = etree.parse(usage_string)
+      except UnicodeError:
+        logger.info("Failed to read %s." % (computer_partition_usage.usage))
+        logger.error(UnicodeError)
+        raise "Failed to read %s." % (computer_partition_usage.usage)
+      except (etree.XMLSyntaxError, etree.DocumentInvalid) as e:
+        logger.info("Failed to parse %s." % (usage_string))
+        logger.error(e)
+        raise _formatXMLError(e)
+      except Exception:
+        raise "Failed to generate XML report."
+
+      for movement in root.findall('movement'):
+        xml_movements += "<movement>"
+        for children in movement.getchildren():
+          if children.tag == "reference":
+            xml_movements += "<%s>%s</%s>" % (children.tag, computer_partition_usage.getId(), children.tag)
+          else:
+            xml_movements += "<%s>%s</%s>" % (children.tag, children.text, children.tag)
+        xml_movements += "</movement>"  
+
+    xml_foot = "</transaction>" \
+               "</journal>"
+
+    xml = xml_head + xml_movements + xml_foot
+    return xml
+
   def agregateAndSendUsage(self):
     """Will agregate usage from each Computer Partition.
     """
@@ -479,9 +710,33 @@ class Slapgrid(object):
     logger = logging.getLogger('UsageReporting')
     logger.info("Aggregating and sending usage reports...")
 
+    #We retrieve XSD models
+    try:
+      computer_consumption_model = \
+        pkg_resources.resource_string(
+          'slapos.slap',
+          'doc/computer_consumption.xsd')
+    except IOError:
+      computer_consumption_model = \
+        pkg_resources.resource_string(
+          __name__, 
+          '../../../../slapos/slap/doc/computer_consumption.xsd')
+
+    try:
+      partition_consumption_model = \
+        pkg_resources.resource_string(
+          'slapos.slap',
+          'doc/partition_consumption.xsd')
+    except IOError:
+      partition_consumption_model = \
+        pkg_resources.resource_string(
+          __name__, 
+          '../../../../slapos/slap/doc/partition_consumption.xsd')
+
     clean_run = True
     #We loop on the different computer partitions
-    for computer_partition in slap_computer_usage.getComputerPartitionList():
+    computer_partition_list = slap_computer_usage.getComputerPartitionList()
+    for computer_partition in computer_partition_list:
       computer_partition_id = computer_partition.getId()
 
       #We want execute all the script in the report folder
@@ -540,7 +795,8 @@ class Slapgrid(object):
 
     #Now we loop through the different computer partitions to ggetId()et reports
     report_usage_issue_cp_list = []
-    for computer_partition in slap_computer_usage.getComputerPartitionList():
+    for computer_partition in computer_partition_list:
+      filename_delete_list = []
       computer_partition_id = computer_partition.getId()
       instance_path = os.path.join(self.instance_root, computer_partition_id)
       dir_reports = os.path.join(instance_path, 'var', 'xml_report')
@@ -558,36 +814,52 @@ class Slapgrid(object):
         file_path = os.path.join(dir_reports, filename)
         if os.path.exists(file_path):
           usage_file = open(file_path, 'r')
-          usage += usage_file.read()
+          usage = usage_file.read()
           usage_file.close()
 
-          computer_partition_usage = self.slap.registerComputerPartition(
-                  self.computer_id, computer_partition_id)
-          computer_partition_usage.setUsage(usage)
-          computer_partition_usage_list.append(computer_partition_usage)
+          #We check the validity of xml content of each reports
+          if not self.validateXML(usage, partition_consumption_model):
+            logger.info('WARNING: The XML file %s generated by slapreport is not valid - ' \
+                            'This report is left as is at %s where you can inspect what went wrong ' % (filename, dir_reports))
+            #Warn the SlapOS Master that a partition generates corrupted xml report
+          else:
+            computer_partition_usage = self.slap.registerComputerPartition(
+                    self.computer_id, computer_partition_id)
+            computer_partition_usage.setUsage(usage)
+            computer_partition_usage_list.append(computer_partition_usage)
+            filename_delete_list.append(filename)
         else:
           logger.debug("Usage report %r not found, ignored" % file_path)
 
-      #-XXX-Here we'll have a "validate_xml" function
-      #if not self.validateXML(usage):
-      ##  logger.debug("WARNING: The file xml %s is not valid " %
-      #       os.path.join(dir_reports, filename))
-      if filename_list:
-        logger.debug("Sending usage informations and terminating for "
-            "partition %s..." % os.path.basename(instance_path))
-        #logger.debug('file %s' % usage)
-      else:
-        logger.debug("No usage file created for partition %s..." \
-                      % os.path.basename(instance_path))
-
         #last_push_date = self.computer.getLastUsagePush()
         #periodicity_timedelta = datetime.timedelta(
         #        self.usage_report_periodicity)
         #if periodicity_timedelta + last_push_date < datetime.datetime.today():
         # Pushes informations, if any
 
+      #After sending the aggregated file we remove all the valid xml reports
+      for filename in filename_delete_list:
+        os.remove(os.path.join(dir_reports, filename))
+
+    for computer_partition_usage in computer_partition_usage_list:
+      logger.info('computer_partition_usage_list : %s - %s' % \
+        (computer_partition_usage.usage, computer_partition_usage.getId()))
+
+    #If there is, at least, one report
+    if computer_partition_usage_list != []:
       try:
-        slap_computer_usage.reportUsage(computer_partition_usage_list)
+        #We generate the final XML report with asXML method
+        computer_consumption = self.asXML(computer_partition_usage_list)
+
+        logger.info('Final xml report : %s' % computer_consumption)
+
+        #We test the XML report before sending it
+        if self.validateXML(computer_consumption, computer_consumption_model):
+          logger.info('XML file generated by asXML is valid')
+          slap_computer_usage.reportUsage(computer_consumption)
+        else:
+          logger.info('XML file generated by asXML is not valid !')
+          raise 'XML file generated by asXML is not valid !'
       except Exception:
         computer_partition_id = computer_partition.getId()
         exception = traceback.format_exc()
@@ -597,11 +869,7 @@ class Slapgrid(object):
         computer_partition.error(issue)
         report_usage_issue_cp_list.append(computer_partition_id)
 
-      #After sending the aggregated file we remove everything in the folder
-      for filename in filename_list:
-        os.remove(os.path.join(dir_reports, filename))
-
-    for computer_partition in slap_computer_usage.getComputerPartitionList():
+    for computer_partition in computer_partition_list:
       computer_partition_id = computer_partition.getId()
       try:
         software_url = computer_partition.getSoftwareRelease().getURI()
@@ -656,6 +924,10 @@ class Slapgrid(object):
           logger.debug('Ignored slap error while trying to inform about '
               'destroying not fully configured Computer Partition %r' %
                   computer_partition.getId())
+        except ServerError as server_error:
+          logger.debug('Ignored server error while trying to inform about '
+              'destroying Computer Partition %r. Error is :\n%r' %
+                  (computer_partition.getId(), server_error.args[0]))
 
     logger.info("Finished usage reports...")
     return clean_run
