@@ -108,6 +108,14 @@ class SoftwareRelease(SlapDocument):
       'url': self._software_release, 
       'computer_id': self._computer_guid})
 
+  def destroyed(self):
+    self._connection_helper.POST('/destroyedSoftwareRelease', {
+      'url': self._software_release, 
+      'computer_id': self._computer_guid})
+
+  def getState(self):
+    return getattr(self, '_requested_state', 'available')
+
 # XXX What is this SoftwareInstance class?
 class SoftwareInstance(SlapDocument):
   """
@@ -139,10 +147,13 @@ class Supply(SlapDocument):
 
   zope.interface.implements(interface.ISupply)
 
-  def supply(self, software_release, computer_guid=None):
+  def supply(self, software_release, computer_guid=None, state='available'):
+    # Note: Zope accepts additional arguments, so state in case of older
+    #       servers will be just ignored.
     self._connection_helper.POST('/supplySupply', {
       'url': software_release,
-      'computer_id': computer_guid})
+      'computer_id': computer_guid,
+      'state': state})
 
 class OpenOrder(SlapDocument):
 
@@ -150,7 +161,7 @@ class OpenOrder(SlapDocument):
 
   def request(self, software_release, partition_reference,
       partition_parameter_kw=None, software_type=None, filter_kw=None,
-      state=None):
+      state=None, shared=False):
     if partition_parameter_kw is None:
       partition_parameter_kw = {}
     if filter_kw is None:
@@ -161,16 +172,25 @@ class OpenOrder(SlapDocument):
         'partition_parameter_xml': xml_marshaller.dumps(partition_parameter_kw),
         'filter_xml': xml_marshaller.dumps(filter_kw),
         'state': xml_marshaller.dumps(state),
+        'shared_xml': xml_marshaller.dumps(shared),
       }
     if software_type is not None:
       request_dict['software_type'] = software_type
-    self._connection_helper.POST('/requestComputerPartition', request_dict)
-    xml = self._connection_helper.response.read()
-    software_instance = xml_marshaller.loads(xml)
-    computer_partition = ComputerPartition(
-      software_instance.slap_computer_id.encode('UTF-8'),
-      software_instance.slap_computer_partition_id.encode('UTF-8'))
-    return computer_partition
+    try:
+      self._connection_helper.POST('/requestComputerPartition', request_dict)
+    except ResourceNotReady:
+      return ComputerPartition(request_dict=request_dict)
+    else:
+      xml = self._connection_helper.response.read()
+      software_instance = xml_marshaller.loads(xml)
+      computer_partition = ComputerPartition(
+        software_instance.slap_computer_id.encode('UTF-8'),
+        software_instance.slap_computer_partition_id.encode('UTF-8'))
+      if shared:
+        computer_partition._synced = True
+        computer_partition._connection_dict = software_instance._connection_dict
+        computer_partition._parameter_dict = software_instance._parameter_dict
+      return computer_partition
 
 def _syncComputerInformation(func):
   """
@@ -211,26 +231,30 @@ class Computer(SlapDocument):
   def getComputerPartitionList(self):
     return [x for x in self._computer_partition_list if x._need_modification]
 
-  def reportUsage(self, computer_partition_list):
-    if computer_partition_list == []:
+  def reportUsage(self, computer_usage):
+    if computer_usage == "":
       return
-    computer = Computer(self._computer_id)
-    computer.computer_partition_usage_list = computer_partition_list
-    marshalled_slap_usage = xml_marshaller.dumps(computer)
     self._connection_helper.POST('/useComputer', {
       'computer_id': self._computer_id,
-      'use_string': marshalled_slap_usage})
+      'use_string': computer_usage})
 
   def updateConfiguration(self, xml):
     self._connection_helper.POST(
         '/loadComputerConfigurationFromXML', { 'xml' : xml })
     return self._connection_helper.response.read()
 
+  def bang(self, message):
+    self._connection_helper.POST('/computerBang', {
+      'computer_id': self._computer_id,
+      'message': message})
+
 def _syncComputerPartitionInformation(func):
   """
   Synchronize computer partition object with server information
   """
   def decorated(self, *args, **kw):
+    if getattr(self, '_synced', 0):
+      return func(self, *args, **kw)
     computer = self._connection_helper.getComputerInformation(self._computer_id)
     found_computer_partition = None
     for computer_partition in computer._computer_partition_list:
@@ -269,13 +293,22 @@ def _syncComputerPartitionInformation(func):
     return func(self, *args, **kw)
   return decorated
 
+
 class ComputerPartition(SlapDocument):
 
   zope.interface.implements(interface.IComputerPartition)
 
-  def __init__(self, computer_id, partition_id):
+  def __init__(self, computer_id=None, partition_id=None, request_dict=None):
+    if request_dict is not None and (computer_id is not None or
+        partition_id is not None):
+      raise TypeError('request_dict conflicts with computer_id and '
+        'partition_id')
+    if request_dict is None and (computer_id is None or partition_id is None):
+      raise TypeError('computer_id and partition_id or request_dict are '
+        'required')
     self._computer_id = computer_id
     self._partition_id = partition_id
+    self._request_dict = request_dict
 
   def __getinitargs__(self):
     return (self._computer_id, self._partition_id, )
@@ -312,13 +345,23 @@ class ComputerPartition(SlapDocument):
         'filter_xml': xml_marshaller.dumps(filter_kw),
         'state': xml_marshaller.dumps(state),
       }
-    self._connection_helper.POST('/requestComputerPartition', request_dict)
-    xml = self._connection_helper.response.read()
-    software_instance = xml_marshaller.loads(xml)
-    computer_partition = ComputerPartition(
-      software_instance.slap_computer_id.encode('UTF-8'),
-      software_instance.slap_computer_partition_id.encode('UTF-8'))
-    return computer_partition
+    try:
+      self._connection_helper.POST('/requestComputerPartition', request_dict)
+    except ResourceNotReady:
+      return ComputerPartition(request_dict=request_dict)
+    else:
+      xml = self._connection_helper.response.read()
+      software_instance = xml_marshaller.loads(xml)
+      computer_partition = ComputerPartition(
+        software_instance.slap_computer_id.encode('UTF-8'),
+        software_instance.slap_computer_partition_id.encode('UTF-8'))
+      if shared:
+        computer_partition._synced = True
+        computer_partition._connection_dict = getattr(software_instance,
+          '_connection_dict', None)
+        computer_partition._parameter_dict = getattr(software_instance,
+          '_parameter_dict', None)
+      return computer_partition
 
   def building(self):
     self._connection_helper.POST('/buildingComputerPartition', {
@@ -354,6 +397,22 @@ class ComputerPartition(SlapDocument):
       'computer_partition_id': self._partition_id,
       'error_log': error_log})
 
+  def bang(self, message):
+    self._connection_helper.POST('/softwareInstanceBang', {
+      'computer_id': self._computer_id,
+      'computer_partition_id': self._partition_id,
+      'message': message})
+
+  def rename(self, new_name, slave_reference=None):
+    post_dict = dict(
+      computer_id=self._computer_id,
+      computer_partition_id=self._partition_id,
+      new_name=new_name,
+    )
+    if slave_reference is not None:
+      post_dict.update(slave_reference=slave_reference)
+    self._connection_helper.POST('/softwareInstanceRename', post_dict)
+
   def getId(self):
     return self._partition_id
 
@@ -376,11 +435,12 @@ class ComputerPartition(SlapDocument):
     else:
       return self._software_release_document
 
-  def setConnectionDict(self, connection_dict):
+  def setConnectionDict(self, connection_dict, slave_reference=None):
     self._connection_helper.POST('/setComputerPartitionConnectionXml', {
       'computer_id': self._computer_id,
       'computer_partition_id': self._partition_id,
-      'connection_xml': xml_marshaller.dumps(connection_dict)})
+      'connection_xml': xml_marshaller.dumps(connection_dict),
+      'slave_reference': slave_reference})
 
   @_syncComputerPartitionInformation
   def getConnectionParameter(self, key):
@@ -416,9 +476,13 @@ class ComputerPartition(SlapDocument):
 #   return decorated 
 
 class ConnectionHelper:
-  error_message_connect_fail = "Couldn't connect to the server. Please double \
-check given master-url argument, and make sure that IPv6 is enabled on \
-your machine and that the server is available. The original error was:"
+  error_message_timeout = "\nThe connection timed out. Please try again later."
+  error_message_connect_fail = "Couldn't connect to the server. Please " \
+      "double check given master-url argument, and make sure that IPv6 is " \
+      "enabled on your machine and that the server is available. The " \
+      "original error was: "
+  ssl_error_message_connect_fail = "\nCouldn't authenticate computer. Please "\
+      "check that certificate and key exist and are valid. "
   def __init__(self, connection_wrapper, host, path, key_file=None,
       cert_file=None, master_ca_file=None, timeout=None):
     self.connection_wrapper = connection_wrapper
@@ -452,6 +516,11 @@ your machine and that the server is available. The original error was:"
         self.connect()
         self.connection.request('GET', self.path + path)
         self.response = self.connection.getresponse()
+      # If ssl error : may come from bad configuration
+      except ssl.SSLError, e:
+        if e.message == "The read operation timed out":
+          raise socket.error(str(e) + self.error_message_timeout)
+        raise ssl.SSLError(str(e) + self.ssl_error_message_connect_fail)
       except socket.error, e:
         raise socket.error(self.error_message_connect_fail + str(e))
       # check self.response.status and raise exception early
@@ -479,6 +548,9 @@ your machine and that the server is available. The original error was:"
         header_dict = {'Content-type': content_type}
         self.connection.request("POST", self.path + path,
             urllib.urlencode(parameter_dict), header_dict)
+      # If ssl error : must come from bad configuration
+      except ssl.SSLError, e:
+        raise ssl.SSLError(self.ssl_error_message_connect_fail + str(e))
       except socket.error, e:
         raise socket.error(self.error_message_connect_fail + str(e))
       self.response = self.connection.getresponse()
